/*
    Copyright 2005-2009 Intel Corporation.  All Rights Reserved.

    This file is part of Threading Building Blocks.

    Threading Building Blocks is free software; you can redistribute it
    and/or modify it under the terms of the GNU General Public License
    version 2 as published by the Free Software Foundation.

    Threading Building Blocks is distributed in the hope that it will be
    useful, but WITHOUT ANY WARRANTY; without even the implied warranty
    of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Threading Building Blocks; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    As a special exception, you may use this file as part of a free software
    library without restriction.  Specifically, if other files instantiate
    templates or use macros or inline functions from this file, or you compile
    this file and link it with other files to produce an executable, this
    file does not by itself cause the resulting executable to be covered by
    the GNU General Public License.  This exception does not however
    invalidate any other reasons why the executable file might be covered by
    the GNU General Public License.
*/

#include "tbb/combinable.h"
#include "tbb/task_scheduler_init.h"
#include "tbb/parallel_for.h"
#include "tbb/parallel_reduce.h"
#include "tbb/blocked_range.h"
#include "tbb/tick_count.h"
#include "tbb/tbb_allocator.h"
#include "tbb/tbb_thread.h"

#include <cstring>
#include <vector>
#include <deque>
#include <list>
#include <map>
#include <utility>

#include "harness_assert.h"
#include "harness.h"
#include <iostream>

static tbb::atomic<int> construction_counter;
static tbb::atomic<int> destruction_counter;

const int REPETITIONS = 10;
const int N = 100000;
const int VALID_NUMBER_OF_KEYS = 100;
const double EXPECTED_SUM = (REPETITIONS + 1) * N;

//
// A minimal class
// Define: default and copy constructor, and allow implicit operator&
// Hide: operator=
//

class minimal: NoAssign {
private:
    int my_value;
public:
    minimal(int val=0) : my_value(val) { ++construction_counter; }
    minimal( const minimal &m ) : my_value(m.my_value) { ++construction_counter; }
    ~minimal() { ++destruction_counter; }
    void set_value( const int i ) { my_value = i; }
    int value( ) const { return my_value; }
};

//
// A helper class that simplifies writing the tests since minimal does not 
// define = or + operators.
//

template< typename T >
struct test_helper {
   static inline void init(T &e) { e = static_cast<T>(0); }  
   static inline void sum(T &e, const int addend ) { e += static_cast<T>(addend); }
   static inline void sum(T &e, const double addend ) { e += static_cast<T>(addend); }
   static inline void set(T &e, const int value ) { e = static_cast<T>(value); }
   static inline double get(const T &e ) { return static_cast<double>(e); }
};

template< >
struct test_helper<minimal> {
   static inline void init(minimal &sum) { sum.set_value( 0 ); }  
   static inline void sum(minimal &sum, const int addend ) { sum.set_value( sum.value() + addend); }
   static inline void sum(minimal &sum, const double addend ) { sum.set_value( sum.value() + static_cast<int>(addend)); }
   static inline void sum(minimal &sum, const minimal &addend ) { sum.set_value( sum.value() + addend.value()); }
   static inline void set(minimal &v, const int value ) { v.set_value( static_cast<int>(value) ); }
   static inline double get(const minimal &sum ) { return static_cast<double>(sum.value()); }
};

//// functors for initialization and combine

// Addition
template <typename T>
struct FunctorAddFinit {
    T operator()() { return 0; }
};

template <>
struct FunctorAddFinit<minimal> {
    minimal operator()() { return minimal(0); }
};

template <typename T>
struct FunctorAddFinit7 {
    T operator()() { return 7; }
};

template <>
struct FunctorAddFinit7<minimal> {
    minimal operator()() { return minimal(7); }
};

template <typename T>
struct FunctorAddCombine {
    T operator()(T left, T right ) const {
        return left + right;
    }
};

template <>
struct FunctorAddCombine<minimal> {
    minimal operator()(minimal left, minimal right ) {
        return minimal(left.value() + right.value()); 
    }
};

template <typename T>
struct FunctorAddCombineRef {
    T operator()(const T& left, const T& right ) const {
        return left + right;
    }
};

template <>
struct FunctorAddCombineRef<minimal> {
    minimal operator()(const minimal& left, const minimal& right ) const {
        return minimal(left.value() + right.value());
    }
};

template <typename T>
T my_finit( ) { return 0; }

template <typename T>
T my_combine( T left, T right) { return left + right; }

template <typename T>
T my_combine_ref( const T &left, const T &right) { return left + right; }

template <>
minimal my_finit( ) { return minimal(0); }

template <>
minimal my_combine( minimal left, minimal right) { return minimal(left.value() + right.value()); }

template <>
minimal my_combine_ref( const minimal &left, const minimal &right) { return minimal(left.value() + right.value()); }

template <typename T>
class combine_each_helper {
public:
    combine_each_helper(T& _result) : my_result(_result) {}
    void operator()(const T& new_bit) { test_helper<T>::sum(my_result, new_bit); }
    combine_each_helper& operator=(const combine_each_helper& other) { 
        test_helper<T>::set(my_result, test_helper<T>::get(other)); 
        return *this; 
    }
private:
    T& my_result;
};

template <typename T>
class combine_each_vector_helper: NoAssign {
public:
    typedef std::vector<T, tbb::tbb_allocator<T> > container_type;
    combine_each_vector_helper(T& _result) : my_result(_result) { }
    // void operator()(const T& new_bit) { test_helper<T>::sum(my_result, new_bit); }
    void operator()(const container_type& new_bit) { 
        for(typename container_type::const_iterator ci = new_bit.begin(); ci != new_bit.end(); ++ci) {
            test_helper<T>::sum(my_result, *ci);
        }
    }
private:
    T& my_result;
};



//// end functors

template< typename T >
void run_serial_scalar_tests(const char *test_name) {
    tbb::tick_count t0;
    T sum;
    test_helper<T>::init(sum);

    if (Verbose) printf("Testing serial %s... ", test_name);  
    for (int t = -1; t < REPETITIONS; ++t) {
        if (Verbose && t == 0) t0 = tbb::tick_count::now(); 
        for (int i = 0; i < N; ++i) {
            test_helper<T>::sum(sum,1); 
        }
    }
 
    double result_value = test_helper<T>::get(sum);
    ASSERT( EXPECTED_SUM == result_value, NULL);
    if (Verbose)
        printf("done\nserial %s, 0, %g, %g\n", test_name, result_value, ( tbb::tick_count::now() - t0).seconds());
}


template <typename T>
class parallel_scalar_body: NoAssign {
    
    tbb::combinable<T> &sums;
 
public:

    parallel_scalar_body ( tbb::combinable<T> &_sums ) : sums(_sums) { }

    void operator()( const tbb::blocked_range<int> &r ) const {
        for (int i = r.begin(); i != r.end(); ++i) { 
            bool was_there;
            T& my_local = sums.local(was_there);
            if(!was_there) test_helper<T>::init(my_local);
            test_helper<T>::sum( my_local, 1 );
        }
    }
   
};

// parallel body with no test for first access.
template <typename T>
class parallel_scalar_body_noinit: NoAssign {
    
    tbb::combinable<T> &sums;
 
public:

    parallel_scalar_body_noinit ( tbb::combinable<T> &_sums ) : sums(_sums) { }

    void operator()( const tbb::blocked_range<int> &r ) const {
        for (int i = r.begin(); i != r.end(); ++i) { 
            test_helper<T>::sum( sums.local(), 1 );
        }
    }
   
};

template< typename T >
void run_parallel_scalar_tests(const char *test_name) {

    tbb::task_scheduler_init init(tbb::task_scheduler_init::deferred);
    T exemplar;
    test_helper<T>::init(exemplar);
    T exemplar23;
    test_helper<T>::set(exemplar23,23);

    for (int p = MinThread; p <= MaxThread; ++p) { 


        if (p == 0) continue;

        if (Verbose) printf("Testing parallel %s on %d thread(s)... ", test_name, p); 
        init.initialize(p);

        tbb::tick_count t0;

        T assign_sum;
        test_helper<T>::init(assign_sum);

        T combine_sum;
        test_helper<T>::init(combine_sum);

        T combine_ref_sum;
        test_helper<T>::init(combine_ref_sum);

        T combine_each_sum;
        test_helper<T>::init(combine_each_sum);

        for (int t = -1; t < REPETITIONS; ++t) {
            if (Verbose && t == 0) t0 = tbb::tick_count::now(); 

            tbb::combinable<T> sums;
            FunctorAddFinit<T> my_finit;
            tbb::combinable<T> finit_combinable(my_finit);
                                    

            tbb::parallel_for( tbb::blocked_range<int>( 0, N, 10000 ), parallel_scalar_body_noinit<T>( finit_combinable ) );
            tbb::parallel_for( tbb::blocked_range<int>( 0, N, 10000 ), parallel_scalar_body<T>( sums ) );

            // Use combine
            test_helper<T>::sum(combine_sum, sums.combine(my_combine<T>));
            test_helper<T>::sum(combine_ref_sum, sums.combine(my_combine_ref<T>));

            combine_each_helper<T> my_helper(combine_each_sum);
            sums.combine_each(my_helper);
           
            // test assignment
            tbb::combinable<T> assigned;
            assigned = sums;

            test_helper<T>::sum(assign_sum, assigned.combine(my_combine<T>));
        }

        ASSERT( EXPECTED_SUM == test_helper<T>::get(combine_sum), NULL);
        ASSERT( EXPECTED_SUM == test_helper<T>::get(combine_ref_sum), NULL);
        ASSERT( EXPECTED_SUM == test_helper<T>::get(assign_sum), NULL);

        if (Verbose)
            printf("done\nparallel %s, %d, %g, %g\n", test_name, p, test_helper<T>::get(combine_sum), 
                                                      ( tbb::tick_count::now() - t0).seconds());
        init.terminate();
    }
}


template <typename T>
class parallel_vector_for_body: NoAssign {
    
    tbb::combinable< std::vector<T, tbb::tbb_allocator<T> > > &locals;
 
public:

    parallel_vector_for_body ( tbb::combinable< std::vector<T, tbb::tbb_allocator<T> > > &_locals ) : locals(_locals) { }

    void operator()( const tbb::blocked_range<int> &r ) const {
        T one;
        test_helper<T>::set(one, 1);

        for (int i = r.begin(); i < r.end(); ++i) {
            locals.local().push_back( one );
        }
    }
   
};

template <typename R, typename T>
struct parallel_vector_reduce_body {

    T sum;    
    size_t count;    

    parallel_vector_reduce_body ( ) : count(0) { test_helper<T>::init(sum); }
    parallel_vector_reduce_body ( parallel_vector_reduce_body<R, T> &, tbb::split ) : count(0) {  test_helper<T>::init(sum); }

    void operator()( const R &r ) {
        for (typename R::iterator ri = r.begin(); ri != r.end(); ++ri) {
            const std::vector< T, tbb::tbb_allocator<T>  > &v = *ri; 
            ++count;
            for (typename std::vector<T, tbb::tbb_allocator<T> >::const_iterator vi = v.begin(); vi != v.end(); ++vi) {
                test_helper<T>::sum(sum, *vi);
            }
        }
    }

    void join( const parallel_vector_reduce_body &b ) {
        test_helper<T>::sum(sum,b.sum);
        count += b.count;
    }
   
};

template< typename T >
void run_parallel_vector_tests(const char *test_name) {
    tbb::tick_count t0;
    tbb::task_scheduler_init init(tbb::task_scheduler_init::deferred);
    typedef std::vector<T, tbb::tbb_allocator<T> > container_type;

    for (int p = MinThread; p <= MaxThread; ++p) { 

        if (p == 0) continue;
        if (Verbose) printf("Testing parallel %s on %d thread(s)... ", test_name, p);
        init.initialize(p);

        T sum;
        T sum2;
        T sum3;
        test_helper<T>::init(sum);
        test_helper<T>::init(sum2);
        test_helper<T>::init(sum3);

        for (int t = -1; t < REPETITIONS; ++t) {
            if (Verbose && t == 0) t0 = tbb::tick_count::now(); 
            typedef typename tbb::combinable< container_type > combinable_type;
            combinable_type vs;

            tbb::parallel_for ( tbb::blocked_range<int> (0, N, 10000), parallel_vector_for_body<T>( vs ) );

            // copy construct
            combinable_type vs2(vs); // this causes an assertion failure, related to allocators...

            // assign
            combinable_type vs3;
            vs3 = vs;

            combine_each_vector_helper<T> my_combine_each(sum);
            vs.combine_each(my_combine_each);

            combine_each_vector_helper<T> my_combine_each2(sum2);
            vs2.combine_each(my_combine_each2);

            combine_each_vector_helper<T> my_combine_each3(sum3);
            vs2.combine_each(my_combine_each3);
            // combine_each sums all elements of each vector into the result.
        }

        double result_value = test_helper<T>::get(sum);
        ASSERT( EXPECTED_SUM == result_value, NULL);
        result_value = test_helper<T>::get(sum2);
        ASSERT( EXPECTED_SUM == result_value, NULL);
        result_value = test_helper<T>::get(sum3);
        ASSERT( EXPECTED_SUM == result_value, NULL);
        if (Verbose)
            printf("done\nparallel %s, %d, %g, %g\n", test_name, p, result_value, ( tbb::tick_count::now() - t0).seconds());
        init.terminate();
    }
}

void 
run_parallel_tests() {
    run_parallel_scalar_tests<int>("int");
    run_parallel_scalar_tests<double>("double");
    run_parallel_scalar_tests<minimal>("minimal");
    run_parallel_vector_tests<int>("std::vector<int, tbb::tbb_allocator<int> >");
    run_parallel_vector_tests<double>("std::vector<double, tbb::tbb_allocator<double> >");
}

template <typename T>
void
run_assign_and_copy_constructor_test(const char *test_name) {
    if (Verbose) printf("Testing assignment and copy construction for %s\n", test_name);

    // test creation with finit function (combine returns finit return value if no threads have created locals)
    FunctorAddFinit7<T> my_finit7;
    tbb::combinable<T> create2(my_finit7);
    ASSERT(7 == test_helper<T>::get(create2.combine(my_combine<T>)), NULL);

    // test copy construction with function initializer
    tbb::combinable<T> copy2(create2);
    ASSERT(7 == test_helper<T>::get(copy2.combine(my_combine<T>)), NULL);

    // test copy assignment with function initializer
    FunctorAddFinit<T> my_finit;
    tbb::combinable<T> assign2(my_finit);
    assign2 = create2;
    ASSERT(7 == test_helper<T>::get(assign2.combine(my_combine<T>)), NULL);
}

void
run_assignment_and_copy_constructor_tests() {
    if(Verbose) printf("Running assignment and copy constructor tests\n");
    run_assign_and_copy_constructor_test<int>("int");
    run_assign_and_copy_constructor_test<double>("double");
    run_assign_and_copy_constructor_test<minimal>("minimal");
}

int main(int argc, char *argv[]) {
   ParseCommandLine(argc, argv);

   if (MaxThread > 0) {
      run_parallel_tests();
   }

    run_assignment_and_copy_constructor_tests();

   printf("done\n");
   return 0;
}

